from json import dumps
from urllib.parse import quote

from .authmixin import HEADER_JSON, AuthError
from .httpclient import HttpClient


class SciCatRestAPI(HttpClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.login_path = self._login_path or "https://dacat.psi.ch/auth/msad"
        self.token_prefix = self.options.get("token_prefix", "") or ""

    def authenticate(self, username, password):
        auth_payload = {"username": username, "password": password}
        res = self._login(auth_payload, HEADER_JSON)
        try:
            token = f"{self.token_prefix}{res['id']}"
        except KeyError as e:
            raise SciCatAuthError(res) from e
        else:
            return token


class SciCatLegacy:
    def __init__(self, *args, **kwargs):
        self.http_client = SciCatRestAPI(*args, **kwargs)

    @property
    def proposals(self):
        url = self.http_client.address + "/proposals"
        return self.http_client.get_request(url, headers=HEADER_JSON)


class SciCatNew:
    max_iterations = 1000

    def __init__(self, *args, return_options=None, **kwargs):
        self.return_options = return_options or {}
        self.http_client = SciCatRestAPI(*args, **kwargs)

    def _proposals_batch(self):
        url = self.http_client.address + "/proposals"
        limit = 500
        filter = {"limits": {"skip": 0, "limit": limit}}
        iteration = 0
        while True:
            iteration += 1
            if iteration > self.max_iterations:
                raise RuntimeError("Exceeded maximum iterations in proposals_batch")
            proposals = self.http_client.get_request(
                f"{url}?filters={quote(dumps(filter))}", headers=HEADER_JSON
            )
            if not proposals or len(proposals) == 0:
                break
            yield proposals
            filter["limits"]["skip"] += limit

    @property
    def proposals(self):
        lazy = self.return_options.get("lazy", False)
        generator = (proposal for batch in self._proposals_batch() for proposal in batch)
        return generator if lazy else list(generator)


class SciCat:
    def __new__(cls, *args, **kwargs):
        token_prefix = kwargs.get("options", {}).get("token_prefix")
        if token_prefix:
            return SciCatNew(*args, **kwargs)
        return SciCatLegacy(*args, **kwargs)


class SciCatAuthError(AuthError):
    pass



#!/usr/bin/env python3

import argparse

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("pgroup", help="Expected form: p12345")
parser.add_argument("-u", "--url", default="https://lnode2.psi.ch/api/v1", help="Server address")

clargs = parser.parse_args()
pgroup = clargs.pgroup
url = clargs.url


import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


from scilog import Basesnippet, Paragraph, SciLog

tmp = Basesnippet()
tmp.id = "2"

log = SciLog(url)
# print(log.token)
logbooks = log.get_logbooks(ownerGroup=pgroup)
print(logbooks)

assert len(logbooks) == 1
logbook = logbooks[0]
print(logbook)

log.select_logbook(logbook)
import time

time.sleep(5)
import datetime

begin_time = datetime.datetime.now()
for ii in range(1000):
    log.send_message(f"<p>from python; number: {ii}</p>")
# print(res)
print(datetime.datetime.now() - begin_time)

# snips = log.get_snippets(snippetType="paragraph", ownerGroup=pgroup)
# print(snips)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, List, Union

from typeguard import typechecked

from .models import Paragraph
from .scilog import SciLogCore

if TYPE_CHECKING:
    from .scilog import SciLog


class LogbookMessage:
    """Class to assist users in composing logbook messages.

    Examples:
        >>> msg = LogbookMessage()
        >>> msg.add_text("First paragraph").add_file("<path_to_the_file>").add_text("Second paragraph below the file")

    """

    def __init__(self, logbook: SciLog | None = None, **kwargs):
        self._content = Paragraph(**kwargs)
        self._logbook = logbook

    @typechecked
    def add_text(self, msg: str):
        """Add text to the paragraph. If the text does not start with
        an html tag '<', it will be automatically embedded in a paragraph <p></p>.

        Args:
            msg (str): Text that ought to be added to the paragraph.

        Returns:
            LogbookMessage: The current logbook message.
        """
        if not msg.startswith("<"):
            msg = f"<p>{msg}</p>"

        if self._content.textcontent:
            self._content.textcontent += msg
        else:
            self._content.textcontent = msg
        return self

    @typechecked
    def add_file(self, file_path: str):
        """Insert a file into the logbook message.

        Args:
            file_path (str): Full path to the file that ought to be added.

        Returns:
            LogbookMessage: The current logbook message.
        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} could not be found.")
        file_info, textcontent = SciLogCore.prepare_file_content(file_path)
        if self._content.files:
            self._content.files.append(file_info)
        else:
            self._content.files = [file_info]
        self.add_text(textcontent)
        return self

    @typechecked
    def add_tag(self, tag: Union[List[str], str]):
        """Add tags to the logbook message.

        Args:
            tag (List[str]): List of tags.

        Returns:
            LogbookMessage: The current logbook message.
        """
        if not isinstance(tag, list):
            tag = [tag]
        if self._content.tags:
            self._content.tags.extend(tag)
        else:
            self._content.tags = tag
        return self

    def send(self):
        """Send the logbook message to the SciLog server. The logbook message must be associated with a logbook instance.

        Raises:
            ValueError: If the logbook message is not associated with a logbook instance.
        """
        if not self._logbook:
            raise ValueError(
                "LogbookMessage must be associated with a logbook instance to be sent."
            )
        self._logbook.send_logbook_message(self)



import json
import os
from unittest import mock

import pytest

from scilog import LogbookMessage, SciLog
from scilog.authmixin import HEADER_JSON
from scilog.models import Filesnippet, Logbook, Paragraph


@pytest.fixture()
def scilog():
    log = SciLog("fake_url")
    log.logbook = Logbook(id="logbook_id_test")
    return log


def test_send_logbook_message(scilog):
    log = scilog
    msg = LogbookMessage()
    msg.add_text("some text")

    with mock.patch.object(log.core, "post_snippet") as mock_post_snippet:
        log.send_logbook_message(msg)
        mock_post_snippet.assert_called_with(
            linkType="paragraph", textcontent="<p>some text</p>", snippetType="paragraph"
        )


def test_send_logbook_message_with_file(scilog, tmpdir):
    log = scilog
    test_file = tmpdir.join("test_file.png")
    test_file.write("test content")
    msg = LogbookMessage()
    msg.add_file(str(test_file))
    msg.add_text("some text")

    with mock.patch.object(log.http_client, "post_request") as post_request:
        post_request.side_effect = [
            {"snippetType": "image", "id": "file1", "accessHash": "hash1"},
            {"snippetType": "paragraph", "id": "para1"},
        ]
        log.send_logbook_message(msg)

    assert post_request.call_count == 2

    file_call = post_request.call_args_list[0]
    snippet_call = post_request.call_args_list[1]

    assert file_call.args[0] == log.http_client.address + "/filesnippet/files"
    assert file_call.kwargs["headers"] == {"Accept": "application/json"}

    assert snippet_call.args[0] == log.http_client.address + "/basesnippets"
    payload = snippet_call.kwargs["payload"]
    assert snippet_call.kwargs["headers"] == HEADER_JSON
    assert payload["linkType"] == "paragraph"
    assert payload["snippetType"] == "paragraph"
    assert "figure" in payload["textcontent"]
    assert "img src" in payload["textcontent"]
    assert payload["files"][0]["fileId"] == "file1"
    assert payload["files"][0]["accessHash"] == "hash1"
    assert "filepath" not in payload["files"][0]


def test_post_snippet(scilog):
    log = scilog
    with mock.patch.object(log.http_client, "post_request") as post_request:
        log.core.post_snippet(textcontent="test")
        post_request.assert_called_with(
            log.http_client.address + "/basesnippets",
            payload={"parentId": "logbook_id_test", "textcontent": "test"},
            headers={"Content-type": "application/json", "Accept": "application/json"},
        )


def test_post_location(scilog):
    log = scilog
    with mock.patch.object(log.http_client, "post_request") as post_request:
        log.core.post_location(name="loc")
        post_request.assert_called_with(
            log.http_client.address + "/basesnippets",
            payload={"parentId": "logbook_id_test", "name": "loc"},
            headers={"Content-type": "application/json", "Accept": "application/json"},
        )


def test_send_logbook_message_data(scilog):
    log = scilog

    msg = LogbookMessage()
    msg.add_text("some text")

    with mock.patch.object(log.http_client, "post_request") as post_request:
        log.send_logbook_message(msg)
        post_request.assert_called_with(
            log.core.http_client.address + "/basesnippets",
            payload={
                "snippetType": "paragraph",
                "parentId": "logbook_id_test",
                "textcontent": "<p>some text</p>",
                "linkType": "paragraph",
            },
            headers={"Content-type": "application/json", "Accept": "application/json"},
        )


def test_send_message_data(scilog):
    log = scilog
    with mock.patch.object(log.http_client, "post_request") as post_request:
        log.send_message("some text")
        post_request.assert_called_with(
            log.core.http_client.address + "/basesnippets",
            payload={
                "snippetType": "paragraph",
                "parentId": "logbook_id_test",
                "textcontent": "<p>some text</p>",
                "linkType": "paragraph",
            },
            headers={"Content-type": "application/json", "Accept": "application/json"},
        )


@pytest.mark.parametrize("filepath,fsnippet", [("./test_file.png", None)])
def test_prepare_file_content_image(scilog, filepath, fsnippet):
    log = scilog
    info, textcontent = log.core.prepare_file_content(filepath, fsnippet)
    assert isinstance(info["fileHash"], str)
    assert len(info["fileHash"]) > 5
    assert info["filepath"] == filepath
    assert info["fileExtension"] == "image/png"
    assert info["fileId"] is None
    assert info["style"] == {"width": "82.25%", "height": ""}

    assert (
        textcontent
        == f'<figure class="image image_resized"><img src="" title="{info["fileHash"]}"></figure>'
    )


@pytest.mark.parametrize("filepath,fsnippet", [("./test_file.pdf", None)])
def test_prepare_file_content_file(scilog, filepath, fsnippet):
    log = scilog
    info, textcontent = log.core.prepare_file_content(filepath, fsnippet)
    assert isinstance(info["fileHash"], str)
    assert len(info["fileHash"]) > 5
    assert info["filepath"] == filepath
    assert info["fileExtension"] == "file/pdf"
    assert info["fileId"] is None

    assert (
        textcontent
        == f'<p><a class="fileLink" target="_blank" href="file:{info["fileHash"]}">{os.path.basename(filepath)}</a></p>'
    )


def test_patch_snippet_updates_and_calls_http(scilog):
    log = scilog
    snippet = Paragraph(
        id="snippet_id",
        parentId="logbook_id_test",
        textcontent="updated",
        createdAt="2024-01-01",
        createdBy="user",
        updatedAt="2024-01-02",
        updatedBy="user",
    )
    response = {"snippetType": "paragraph", "id": "updated_id", "textcontent": "updated"}

    with mock.patch.object(
        log.http_client, "patch_request", return_value=response
    ) as patch_request:
        result = log.core.patch_snippet(snippet)

    assert isinstance(result, Paragraph)
    patch_request.assert_called_once()
    (url,) = patch_request.call_args.args
    payload = patch_request.call_args.kwargs["payload"]
    headers = patch_request.call_args.kwargs["headers"]

    assert url == log.http_client.address + "/basesnippets/snippet_id"
    assert headers == HEADER_JSON
    assert payload["snippetType"] == "paragraph"
    assert payload["parentId"] == "logbook_id_test"
    assert payload["textcontent"] == "updated"
    assert "id" not in payload
    assert "createdAt" not in payload
    assert "createdBy" not in payload
    assert "updatedAt" not in payload
    assert "updatedBy" not in payload


def test_append_files_to_snippet_calls_patch(scilog):
    log = scilog
    snippet = Paragraph(textcontent="start")
    fsnippet = Filesnippet(id="file1", accessHash="hash1")

    with mock.patch.object(log.core, "post_file", return_value=fsnippet) as post_file:
        with mock.patch.object(log.core, "patch_snippet", return_value=snippet) as patch_snippet:
            result = log.core.append_files_to_snippet(snippet, ["/tmp/test.png"])

    assert result is snippet
    post_file.assert_called_once_with("/tmp/test.png")
    patch_snippet.assert_called_once_with(snippet)
    assert isinstance(snippet.files, list)
    assert snippet.files[0]["fileId"] == "file1"
    assert snippet.files[0]["accessHash"] == "hash1"
    assert "figure" in snippet.textcontent


def test_get_logbooks_with_kwargs_builds_filter_and_calls_http(scilog):
    log = scilog
    response = [
        {"snippetType": "logbook", "id": "log1", "name": "Log 1"},
        {"snippetType": "logbook", "id": "log2", "name": "Log 2"},
    ]

    with mock.patch.object(log.http_client, "get_request", return_value=response) as get_request:
        result = log.get_logbooks(name="test")

    assert isinstance(result, list)
    assert all(isinstance(item, Logbook) for item in result)

    get_request.assert_called_once()
    (url,) = get_request.call_args.args
    params = get_request.call_args.kwargs["params"]
    headers = get_request.call_args.kwargs["headers"]

    assert url == log.http_client.address + "/logbooks"
    assert headers == HEADER_JSON
    filt = json.loads(params["filter"])
    assert filt == {"where": {"and": [{"name": "test"}]}}


def test_get_logbooks_with_where_calls_http(scilog):
    log = scilog
    response = {"snippetType": "logbook", "id": "log1", "name": "Log 1"}
    where = {"name": "test"}

    with mock.patch.object(log.http_client, "get_request", return_value=response) as get_request:
        result = log.get_logbooks(where=where, limit=2)

    assert isinstance(result, Logbook)

    get_request.assert_called_once()
    (url,) = get_request.call_args.args
    params = get_request.call_args.kwargs["params"]
    headers = get_request.call_args.kwargs["headers"]

    assert url == log.http_client.address + "/logbooks"
    assert headers == HEADER_JSON
    filt = json.loads(params["filter"])
    assert filt == {"where": where, "limit": 2}


def test_get_snippets_with_kwargs_calls_http(scilog):
    log = scilog
    response = {"snippetType": "paragraph", "id": "p1", "textcontent": "text"}

    with mock.patch.object(log.http_client, "get_request", return_value=response) as get_request:
        result = log.get_snippets(textcontent="text")

    assert isinstance(result, Paragraph)

    get_request.assert_called_once()
    (url,) = get_request.call_args.args
    params = get_request.call_args.kwargs["params"]
    headers = get_request.call_args.kwargs["headers"]

    assert url == log.http_client.address + "/basesnippets"
    assert headers == HEADER_JSON
    filt = json.loads(params["filter"])
    assert filt == {"where": {"and": [{"textcontent": "text", "parentId": "logbook_id_test"}]}}


def test_scilog_without_auto_save_does_not_use_config(scilog):
    log = SciLog("fake_url", options={"auto_save_token": False})
    assert log.http_client._auto_save_token is False
    with mock.patch.object(
        log.core.http_client, "_authenticate", return_value="dummy_token"
    ) as mock_authenticate:
        with mock.patch.object(log.core.http_client, "config") as mock_config:
            token = log.http_client.token
        mock_authenticate.assert_called_once()

        # Ensure that the config is neither read from nor written to when auto_save_token is False
        mock_config.__getitem__.assert_not_called()
        mock_config.__setitem__.assert_not_called()
        assert token == "dummy_token"



from __future__ import annotations

from typing import Self

from pydantic import BaseModel, Field


class Basesnippet(BaseModel):
    snippetType: str = "basesnippet"
    id: str | None = None
    parentId: str | None = None
    createACL: list[str] | None = None
    readACL: list[str] | None = None
    updateACL: list[str] | None = None
    deleteACL: list[str] | None = None
    shareACL: list[str] | None = None
    adminACL: list[str] | None = None
    isPrivate: bool | None = None
    createdAt: str | None = None
    createdBy: str | None = None
    updatedAt: str | None = None
    updatedBy: str | None = None
    expiresAt: str | None = None
    subsnippets: list | None = None  # what type?
    tags: list[str] | None = None
    dashboardName: str | None = None
    files: list | None = None
    location: str | None = None
    defaultOrder: int | None = None
    linkType: str | None = None
    versionable: bool | None = None
    deleted: bool | None = None
    ownerGroup: str | None = Field(default=None, deprecated=True)
    accessGroups: list | None = Field(default=None, deprecated=True)

    @classmethod
    def from_http_response(cls, response: dict | list[dict]) -> Self | list[Self]:
        """
        Create an instance of the class from an HTTP response.

        Args:
            response (dict | list[dict]): The HTTP response data.

        Returns:
            Basesnippet | list[Basesnippet]: An instance or list of instances of the class.
        """
        if isinstance(response, list):
            return [cls(**resp) for resp in response]
        return cls(**response)


class Location(Basesnippet):
    snippetType: str = "location"
    name: str | None = None


class Logbook(Basesnippet):
    snippetType: str = "logbook"
    name: str | None = None
    description: str | None = None
    thumbnail: str | None = None


class Paragraph(Basesnippet):
    snippetType: str = "paragraph"
    textcontent: str | None = None
    isMessage: str | None = None


class Filesnippet(Basesnippet):
    snippetType: str = "image"
    fileExtension: str | None = None
    accessHash: str | None = None


def snippet_factory(response):
    """
    Factory function to create the appropriate snippet instance based on the snippetType in the response.

    Args:
        response (dict | list[dict]): The HTTP response data.

    Returns:
        Basesnippet | list[Basesnippet]: An instance or list of instances of the appropriate snippet class.
    """
    if isinstance(response, list):
        return [get_snippet_response(resp) for resp in response]
    return get_snippet_response(response)


def get_snippet_response(response: dict) -> Basesnippet:
    """
    Get the appropriate snippet instance based on the snippetType in the response.

    Args:
        response (dict): The HTTP response data.

    Returns:
        Basesnippet: An instance of the appropriate snippet class.
    """
    available_snippet_types = {
        "location": Location,
        "logbook": Logbook,
        "paragraph": Paragraph,
        "image": Filesnippet,
        "basesnippet": Basesnippet,
    }
    snippet_type = response["snippetType"]
    if snippet_type in available_snippet_types:
        return available_snippet_types[snippet_type].from_http_response(response)
    raise ValueError(f"Unknown snippet type: {snippet_type}")



from unittest.mock import ANY, Mock, patch

import pytest

from scilog import SciCat


@pytest.mark.parametrize(
    "token_prefix, expected_class",
    [
        ["", "SciCatLegacy"],
        [None, "SciCatLegacy"],
        ["Bearer ", "SciCatNew"],
    ],
)
def test_new(token_prefix, expected_class):
    scicat = SciCat(
        "http://scicat",
        options={
            "token_prefix": token_prefix,
        },
    )
    assert scicat.__class__.__name__ == expected_class
